/*
 * This file is part of ZigBrains.
 *
 * Copyright (C) 2023-2025 FalsePattern
 * All Rights Reserved
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * ZigBrains is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, only version 3 of the License.
 *
 * ZigBrains is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with ZigBrains. If not, see <https://www.gnu.org/licenses/>.
 */

package com.falsepattern.zigbrains.project.steps.ui

import com.falsepattern.zigbrains.ZigBrainsBundle
import com.falsepattern.zigbrains.project.execution.build.ZigConfigTypeBuild
import com.falsepattern.zigbrains.project.execution.build.ZigExecConfigBuild
import com.falsepattern.zigbrains.project.execution.firstConfigFactory
import com.falsepattern.zigbrains.project.steps.discovery.ZigStepDiscoveryListener
import com.falsepattern.zigbrains.project.steps.discovery.zigStepDiscovery
import com.falsepattern.zigbrains.shared.coroutine.withEDTContext
import com.falsepattern.zigbrains.shared.ipc.IPCUtil
import com.falsepattern.zigbrains.shared.ipc.ZigIPCService
import com.falsepattern.zigbrains.shared.ipc.ipc
import com.falsepattern.zigbrains.shared.zigCoroutineScope
import com.intellij.execution.ProgramRunnerUtil
import com.intellij.execution.RunManager
import com.intellij.execution.RunnerAndConfigurationSettings
import com.intellij.execution.executors.DefaultRunExecutor
import com.intellij.icons.AllIcons
import com.intellij.openapi.Disposable
import com.intellij.openapi.actionSystem.ActionManager
import com.intellij.openapi.actionSystem.DefaultActionGroup
import com.intellij.openapi.application.ModalityState
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.SimpleToolWindowPanel
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.wm.ToolWindow
import com.intellij.ui.AnimatedIcon
import com.intellij.ui.components.JBLabel
import com.intellij.ui.components.JBScrollPane
import com.intellij.ui.components.JBTextArea
import com.intellij.ui.components.panels.VerticalLayout
import com.intellij.ui.content.Content
import com.intellij.ui.content.ContentFactory
import com.intellij.ui.treeStructure.Tree
import kotlinx.coroutines.launch
import kotlinx.coroutines.sync.withLock
import kotlinx.coroutines.withTimeoutOrNull
import java.awt.Component
import java.awt.GridBagConstraints
import java.awt.GridBagLayout
import java.awt.event.MouseAdapter
import java.awt.event.MouseEvent
import java.util.concurrent.atomic.AtomicBoolean
import javax.swing.BoxLayout
import javax.swing.JPanel
import javax.swing.SwingConstants
import javax.swing.tree.DefaultMutableTreeNode
import javax.swing.tree.DefaultTreeModel
import javax.swing.tree.MutableTreeNode
import javax.swing.tree.TreePath


@OptIn(ExperimentalUnsignedTypes::class)
class BuildToolWindowContext(private val project: Project): Disposable {
    inner class TreeBox() {
        val panel = JPanel(VerticalLayout(0))
        val root = DefaultMutableTreeNode(BaseNodeDescriptor<Any>(project, ""))
        val model = DefaultTreeModel(root)
        val tree = Tree(model).also { it.isRootVisible = false }
    }
    private val viewPanel = JPanel(VerticalLayout(0))
    private val steps = TreeBox()
    private val build = if (IPCUtil.haveIPC) TreeBox() else null
    private var live = AtomicBoolean(true)
    init {
        viewPanel.add(JBLabel(ZigBrainsBundle.message("build.tool.window.tree.steps.label")))
        viewPanel.add(steps.panel)
        steps.panel.setNotScanned()

        steps.tree.addMouseListener(object : MouseAdapter() {
            override fun mouseClicked(e: MouseEvent) {
                if (e.clickCount == 2) {
                    val node = steps.tree.lastSelectedPathComponent as? DefaultMutableTreeNode ?: return
                    val step = node.userObject as? StepNodeDescriptor ?: return

                    val stepName = step.element?.name ?: return
                    val manager = RunManager.getInstance(project)
                    val config = getExistingRunConfig(manager, stepName) ?: run {
                        val factory = firstConfigFactory<ZigConfigTypeBuild>()
                        val newConfig = manager.createConfiguration("zig build $stepName", factory)
                        val config = newConfig.configuration as ZigExecConfigBuild
                        config.buildSteps.args = stepName
                        manager.addConfiguration(newConfig)
                        return@run newConfig
                    }

                    manager.selectedConfiguration = config
                    ProgramRunnerUtil.executeConfiguration(config, DefaultRunExecutor.getRunExecutorInstance())
                }
            }
        })

        if (build != null) {
            viewPanel.add(JBLabel(ZigBrainsBundle.message("build.tool.window.tree.build.label")))
            viewPanel.add(build.panel)
            build.panel.setNoBuilds()

            project.zigCoroutineScope.launch {
                while (!project.isDisposed && live.get()) {
                    val ipc = project.ipc ?: return@launch
                    withTimeoutOrNull(1000) {
                        ipc.changed.receive()
                    } ?: continue
                    ipc.mutex.withLock {
                        withEDTContext(ModalityState.any()) {
                            if (ipc.nodes.isEmpty()) {
                                build.root.removeAllChildren()
                                build.panel.setNoBuilds()
                                return@withEDTContext
                            }
                            val allNodes = ArrayList(ipc.nodes)
                            val existingNodes = ArrayList<ZigIPCService.IPCTreeNode>()
                            val removedNodes = ArrayList<ZigIPCService.IPCTreeNode>()
                            build.root.children().iterator().forEach { child ->
                                if (child !is ZigIPCService.IPCTreeNode) {
                                    return@forEach
                                }
                                if (child !in allNodes) {
                                    removedNodes.add(child)
                                } else {
                                    existingNodes.add(child)
                                }
                            }
                            val newNodes = ArrayList<MutableTreeNode>(allNodes)
                            newNodes.removeAll(existingNodes)
                            removedNodes.forEach { build.root.remove(it) }
                            newNodes.forEach { build.root.add(it) }
                            if (removedNodes.isNotEmpty() || newNodes.isNotEmpty()) {
                                build.model.reload(build.root)
                            }
                            if (build.root.childCount == 0) {
                                build.panel.setNoBuilds()
                            } else {
                                build.panel.setViewportBody(build.tree)
                            }
                            for (bn in allNodes) {
                                expandRecursively(build, bn)
                            }
                        }
                    }
                }
            }
        }
    }

    private fun createContentPanel(): Content {
        val contentPanel = SimpleToolWindowPanel(false)
        val body = JPanel(GridBagLayout())
        contentPanel.setLayout(GridBagLayout())
        val c = GridBagConstraints()
        c.fill = GridBagConstraints.BOTH
        c.weightx = 1.0
        c.weighty = 1.0
        contentPanel.add(body, c)
        c.fill = GridBagConstraints.HORIZONTAL
        c.anchor = GridBagConstraints.NORTHWEST
        c.weightx = 1.0
        c.weighty = 0.0
        val toolbar = ActionManager.getInstance().createActionToolbar("ZigToolbar", DefaultActionGroup(ActionManager.getInstance().getAction("zigbrains.discover.steps")), true)
        toolbar.targetComponent = null
        body.add(toolbar.component, c)
        c.gridwidth = 1
        c.gridy = 1
        c.weighty = 1.0
        c.fill = GridBagConstraints.BOTH
        val viewport = JBScrollPane()
        viewport.setViewportView(viewPanel)
        body.add(viewport, c)
        val content = ContentFactory.getInstance().createContent(contentPanel, "", false)
        return content
    }

    override fun dispose() {
        live.set(false)
    }

    companion object {
        suspend fun create(project: Project, window: ToolWindow) {
            withEDTContext(ModalityState.any()) {
                val context = BuildToolWindowContext(project)
                Disposer.register(context, project.zigStepDiscovery.register(context.BuildReloadListener()))
                Disposer.register(window.disposable, context)
                window.contentManager.addContent(context.createContentPanel())
            }
        }

        private fun expandRecursively(box: TreeBox, node: ZigIPCService.IPCTreeNode) {
            if (node.changed) {
                box.model.reload(node)
                node.changed = false
            }
            box.tree.expandPath(TreePath(box.model.getPathToRoot(node)))
            node.children().asIterator().forEach { child ->
                (child as? ZigIPCService.IPCTreeNode)?.let { expandRecursively(box, it) }
            }
        }
    }

    inner class BuildReloadListener: ZigStepDiscoveryListener {
        override suspend fun preReload() {
            steps.panel.setRunningZigBuild()
        }

        override suspend fun postReload(stepInfo: List<Pair<String, String?>>) {
            steps.root.removeAllChildren()
            for ((task, description) in stepInfo) {
                val icon = when(task) {
                    "install" -> AllIcons.Actions.Install
                    "uninstall" -> AllIcons.Actions.Uninstall
                    else -> AllIcons.RunConfigurations.TestState.Run
                }
                steps.root.add(DefaultMutableTreeNode(StepNodeDescriptor(project, task, icon, description)))
            }
            withEDTContext(ModalityState.any()) {
                steps.model.reload(steps.root)
                steps.panel.setViewportBody(steps.tree)
            }
        }

        override suspend fun errorReload(type: ZigStepDiscoveryListener.ErrorType, details: String?) {
            withEDTContext(ModalityState.any()) {
                steps.panel.setViewportError(ZigBrainsBundle.message(when(type) {
                    ZigStepDiscoveryListener.ErrorType.MissingToolchain -> "build.tool.window.status.error.missing-toolchain"
                    ZigStepDiscoveryListener.ErrorType.MissingZigExe -> "build.tool.window.status.error.missing-zig-exe"
                    ZigStepDiscoveryListener.ErrorType.MissingBuildZig -> "build.tool.window.status.error.missing-build-zig"
                    ZigStepDiscoveryListener.ErrorType.GeneralError -> "build.tool.window.status.error.general"
                }), details)
            }
        }

        override suspend fun timeoutReload(seconds: Int) {
            withEDTContext(ModalityState.any()) {
                steps.panel.setViewportError(ZigBrainsBundle.message("build.tool.window.status.timeout", seconds), null)
            }
        }
    }
}

private fun JPanel.setViewportBody(component: Component) {
    removeAll()
    add(component)
    repaint()
}

private fun JPanel.setRunningZigBuild() {
    setViewportBody(JBLabel(ZigBrainsBundle.message("build.tool.window.status.loading"), AnimatedIcon.Default(), SwingConstants.CENTER))
}

private fun JPanel.setNotScanned() {
    setViewportBody(JBLabel(ZigBrainsBundle.message("build.tool.window.status.not-scanned"), AllIcons.General.Information, SwingConstants.CENTER))
}

private fun JPanel.setNoBuilds() {
    setViewportBody(JBLabel(ZigBrainsBundle.message("build.tool.window.status.no-builds"), AllIcons.General.Information, SwingConstants.CENTER))
}

private fun JPanel.setViewportError(msg: String, details: String?) {
    val result = JPanel()
    result.layout = BoxLayout(result, BoxLayout.Y_AXIS)
    result.add(JBLabel(msg, AllIcons.General.Error, SwingConstants.CENTER))
    if (details != null) {
        val code = JBTextArea()
        code.isEditable = false
        code.text = details
        val scroll = JBScrollPane(code)
        result.add(scroll)
    }
    setViewportBody(result)
}

private fun getExistingRunConfig(manager: RunManager, stepName: String): RunnerAndConfigurationSettings? {
    for (config in manager.getConfigurationSettingsList(ZigConfigTypeBuild::class.java)) {
        val build = config.configuration as? ZigExecConfigBuild ?: continue
        val steps = build.buildSteps.argsSplit()
        if (steps.size != 1)
            continue
        if (steps[0] != stepName)
            continue
        return config
    }
    return null
}
